#ifndef ZP_OPTIMIZER_H
#define ZP_OPTIMIZER_H

#include "Parser.h"

#include <unordered_map>

namespace zp
{
    // forward declarations
    enum class Instruction : char;

    struct Block;
    struct InstructionBlock;
    struct UnionBlock;
    struct Loop;

    class AstVisitor;

    struct BlockInstruction;
    struct SimpleBlockInstruction;
    struct LinearLoopInstruction;
    struct LoopInstruction;

    class Optimizer;

    using Index = int;
    using Cell = unsigned char;
}


class zp::AstVisitor
{
public:
    std::unique_ptr<BlockInstruction> visit(Block& block);

    virtual std::unique_ptr<BlockInstruction>
        visitLoop(Loop& loop) = 0;
    virtual std::unique_ptr<BlockInstruction>
        visitInstructionBlock(InstructionBlock& ib) = 0;
    virtual std::unique_ptr<BlockInstruction>
        visitUnionBlock(UnionBlock& ub) = 0;
};


struct zp::BlockInstruction
{
    virtual ~BlockInstruction(void) = default;
};


struct zp::SimpleBlockInstruction :
    public BlockInstruction
{
    Index pointerMove;
    std::unordered_map<Index, Cell> added;
};


struct zp::LinearLoopInstruction :
    public BlockInstruction
{
    
};


struct zp::LoopInstruction :
    public BlockInstruction
{
    std::vector<std::unique_ptr<BlockInstruction>> content;
};


class zp::Optimizer :
    public AstVisitor
{
public:
    std::unique_ptr<BlockInstruction> visitLoop(Loop& loop) override;
    std::unique_ptr<BlockInstruction> visitInstructionBlock(InstructionBlock& ib) override;
    std::unique_ptr<BlockInstruction> visitUnionBlock(UnionBlock& ub) override;

};


#endif /* ZP_OPTIMIZER_H */
